import graphene
from addict import Dict
from django.test import TestCase
from graphene import Node, Schema
from graphql_relay import to_global_id

from django_scoped_permissions.graphql import (
    ScopedDjangoNode,
    ScopedDjangoCreateMutation, ScopedDjangoUpdateMutation, ScopedDjangoPatchMutation,
)
from django_scoped_permissions.guards import ScopedPermissionGuard
from django_scoped_permissions.tests.factories import UserFactory
from django_scoped_permissions.tests.models import User


class TestScopedDjangoNode(TestCase):
    def test__node_without_explicit_permissions__has_default_permission_handling(
        self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        class UserNode(ScopedDjangoNode):
            class Meta:
                model = User

        class Query(graphene.ObjectType):
            user = Node.Field(UserNode)

        user = UserFactory.create(first_name="Tormod", last_name="Haugland")
        user_two = UserFactory.create()

        schema = Schema(query=Query)
        query = """
            query User($id: ID!){
                user(id: $id){
                    id
                    firstName
                    lastName 
                } 
            }
        """

        result = schema.execute(
            query,
            variables={"id": to_global_id("UserNode", user.id)},
            context=Dict(user=user),
        )
        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.user.firstName)

        # User two does not have permission by default
        result = schema.execute(
            query,
            variables={"id": to_global_id("UserNode", user.id)},
            context=Dict(user=user_two),
        )
        self.assertIsNotNone(result.errors)

        # Can be accessed by someone who has read permissions
        user_two.add_or_create_permission("user:read")
        result = schema.execute(
            query,
            variables={"id": to_global_id("UserNode", user.id)},
            context=Dict(user=user),
        )
        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.user.firstName)

    def test__node_with_explicit_permissions__has_explicit_permission_handling(
        self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        class UserNode(ScopedDjangoNode):
            class Meta:
                model = User
                node_permissions = ("user:read",)

        class Query(graphene.ObjectType):
            user = Node.Field(UserNode)

        user = UserFactory.create(first_name="Tormod", last_name="Haugland")
        user.add_or_create_permission("user:read")
        user_two = UserFactory.create()

        schema = Schema(query=Query)
        query = """
            query User($id: ID!){
                user(id: $id){
                    id
                    firstName
                    lastName 
                } 
            }
        """

        result = schema.execute(
            query,
            variables={"id": to_global_id("UserNode", user.id)},
            context=Dict(user=user),
        )
        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.user.firstName)

        # User two does not have permission by default
        result = schema.execute(
            query,
            variables={"id": to_global_id("UserNode", user.id)},
            context=Dict(user=user_two),
        )
        self.assertIsNotNone(result.errors)

    def test__node_with_permission_guards__has_explicit_permission_handling(
        self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        class UserNode(ScopedDjangoNode):
            class Meta:
                model = User
                node_permissions = ScopedPermissionGuard(scope="user", verb="read")

        class Query(graphene.ObjectType):
            user = Node.Field(UserNode)

        user = UserFactory.create(first_name="Tormod", last_name="Haugland")
        user.add_or_create_permission("user:read")
        user_two = UserFactory.create()

        schema = Schema(query=Query)
        query = """
            query User($id: ID!){
                user(id: $id){
                    id
                    firstName
                    lastName 
                } 
            }
        """

        result = schema.execute(
            query,
            variables={"id": to_global_id("UserNode", user.id)},
            context=Dict(user=user),
        )
        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.user.firstName)

        # User two does not have permission by default
        result = schema.execute(
            query,
            variables={"id": to_global_id("UserNode", user.id)},
            context=Dict(user=user_two),
        )
        self.assertIsNotNone(result.errors)


class TestScopedCreateMutation(TestCase):
    def test__permissions_set__respects_permissions(
        self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        from .schema import UserNode

        class CreateUserMutation(ScopedDjangoCreateMutation):
            class Meta:
                model = User
                exclude_fields = ("password",)
                permissions = ("user:create",)

        class Mutations(graphene.ObjectType):
            create_user = CreateUserMutation.Field()

        user = UserFactory.create()
        user.add_or_create_permission("user:create")

        schema = Schema(mutation=Mutations)
        mutation = """
            mutation CreateUser(
                $input: CreateUserInput! 
            ){
                createUser(input: $input){
                    user{
                        id
                        firstName
                        lastName
                        email
                    }
                }
            }
        """

        result = schema.execute(
            mutation,
            variables={
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user),
        )

        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.createUser.user.firstName)
        self.assertEqual("Haugland", data.createUser.user.lastName)
        self.assertEqual("tormod.haugland@gmail.com", data.createUser.user.email)

    def test__set_permission_via_guard__respects_permission(
        self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        from .schema import UserNode

        class CreateUserMutation(ScopedDjangoCreateMutation):
            class Meta:
                model = User
                exclude_fields = ("password",)
                permissions = ScopedPermissionGuard("user:create")

        class Mutations(graphene.ObjectType):
            create_user = CreateUserMutation.Field()

        user = UserFactory.create()
        user.add_or_create_permission("user:create")

        schema = Schema(mutation=Mutations)
        mutation = """
            mutation CreateUser(
                $input: CreateUserInput! 
            ){
                createUser(input: $input){
                    user{
                        id
                        firstName
                        lastName
                        email
                    }
                }
            }
        """

        result = schema.execute(
            mutation,
            variables={
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user),
        )

        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.createUser.user.firstName)
        self.assertEqual("Haugland", data.createUser.user.lastName)
        self.assertEqual("tormod.haugland@gmail.com", data.createUser.user.email)


class TestScopedUpdateMutation(TestCase):
    def test__default_permission__uses_object_required_scopes(
            self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        from .schema import UserNode

        class UpdateUserMutation(ScopedDjangoUpdateMutation):
            class Meta:
                model = User
                exclude_fields = ("password",)

        class Mutations(graphene.ObjectType):
            update_user = UpdateUserMutation.Field()

        user = UserFactory.create()
        user_two = UserFactory.create()

        schema = Schema(mutation=Mutations)
        mutation = """
            mutation UpdateUser(
                $id: ID!,
                $input: UpdateUserInput! 
            ){
                updateUser(id: $id, input: $input){
                    user{
                        id
                        firstName
                        lastName
                        email
                    }
                }
            }
        """

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user),
        )

        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.updateUser.user.firstName)
        self.assertEqual("Haugland", data.updateUser.user.lastName)
        self.assertEqual("tormod.haugland@gmail.com", data.updateUser.user.email)

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user_two),
        )

        self.assertIsNotNone(result.errors)

    def test__specific_permission__respects_permission(
            self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        from .schema import UserNode

        class UpdateUserMutation(ScopedDjangoUpdateMutation):
            class Meta:
                model = User
                exclude_fields = ("password",)
                permissions = ("user:update",)

        class Mutations(graphene.ObjectType):
            update_user = UpdateUserMutation.Field()

        user = UserFactory.create()
        user.add_or_create_permission("user:update")
        user_two = UserFactory.create()

        schema = Schema(mutation=Mutations)
        mutation = """
            mutation UpdateUser(
                $id: ID!,
                $input: UpdateUserInput! 
            ){
                updateUser(id: $id, input: $input){
                    user{
                        id
                        firstName
                        lastName
                        email
                    }
                }
            }
        """

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user),
        )

        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.updateUser.user.firstName)
        self.assertEqual("Haugland", data.updateUser.user.lastName)
        self.assertEqual("tormod.haugland@gmail.com", data.updateUser.user.email)

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user_two),
        )

        self.assertIsNotNone(result.errors)


class TestScopedPatchMutation(TestCase):
    def test__default_permission__uses_object_required_scopes(
            self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        from .schema import UserNode

        class PatchUserMutation(ScopedDjangoPatchMutation):
            class Meta:
                model = User
                exclude_fields = ("password",)

        class Mutations(graphene.ObjectType):
            patch_user = PatchUserMutation.Field()

        user = UserFactory.create()
        user_two = UserFactory.create()

        schema = Schema(mutation=Mutations)
        mutation = """
            mutation PatchUser(
                $id: ID!,
                $input: PatchUserInput! 
            ){
                patchUser(id: $id, input: $input){
                    user{
                        id
                        firstName
                        lastName
                        email
                    }
                }
            }
        """

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user),
        )

        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.patchUser.user.firstName)
        self.assertEqual("Haugland", data.patchUser.user.lastName)
        self.assertEqual("tormod.haugland@gmail.com", data.patchUser.user.email)

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user_two),
        )

        self.assertIsNotNone(result.errors)

    def test__specific_permission__respects_permission(
            self,
    ):
        # This registers the UserNode type
        # noinspection PyUnresolvedReferences
        from .schema import UserNode

        class PatchUserMutation(ScopedDjangoPatchMutation):
            class Meta:
                model = User
                exclude_fields = ("password",)
                permissions = ("user:patch",)

        class Mutations(graphene.ObjectType):
            patch_user = PatchUserMutation.Field()

        user = UserFactory.create()
        user.add_or_create_permission("user:patch")
        user_two = UserFactory.create()

        schema = Schema(mutation=Mutations)
        mutation = """
            mutation PatchUser(
                $id: ID!,
                $input: PatchUserInput! 
            ){
                patchUser(id: $id, input: $input){
                    user{
                        id
                        firstName
                        lastName
                        email
                    }
                }
            }
        """

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user),
        )

        self.assertIsNone(result.errors)
        data = Dict(result.data)
        self.assertEqual("Tormod", data.patchUser.user.firstName)
        self.assertEqual("Haugland", data.patchUser.user.lastName)
        self.assertEqual("tormod.haugland@gmail.com", data.patchUser.user.email)

        result = schema.execute(
            mutation,
            variables={
                "id": to_global_id("UserNode", user.id),
                "input": {
                    "username": "tormod",
                    "firstName": "Tormod",
                    "lastName": "Haugland",
                    "email": "tormod.haugland@gmail.com",
                }
            },
            context=Dict(user=user_two),
        )

        self.assertIsNotNone(result.errors)



from typing import Any, Union, Optional
from django.db.models import Model
from django.db.models.base import ModelBase


def scope_grants_permission(
    required_scope: str, granting_scope: str, verb: Optional[str] = None
):
    """
    scope_grants_permission checks if a single granting scope matches a single required scope.

    This is slightly simpler than the method scopes_grant_permission below, and follows the following rules:

    1. If the granting_scope has an exclusion rule (starts with "-"), we always return false.
    2. If the scopes are equal, we always returns true.
    3. If the granting_scope is an exact rule (starts with "="), we expand the required scope non-recursively with the verb, and then check for equality.
    4. Exapnd the required scope recursively with the verb and check if the scopes match.

    """
    # Single negation permissions will never grant access
    if granting_scope.startswith("-"):
        return False

    if required_scope == granting_scope:
        return True

    # The equal case will not have the verb applied recursively
    if granting_scope.startswith("="):
        expanded_scopes = expand_scopes_with_verb([required_scope], verb)
    else:
        expanded_scopes = expand_scopes_with_verb_recursively([required_scope], verb)

    return any_scope_matches(expanded_scopes, [granting_scope])


def scopes_grant_permissions(
    required_scopes: [str], granting_scopes: [str], verb: Optional[str] = None
):
    """
    scopes_grants_permission takes as arguments a number of required base scopes,
    and then a set of granting scopes, e.g. scopes a User has.
    The method then checks whether or not the granting scopes provides access to the
    required base scopes.
    This depends on multiple facts. But primarily, we check the following:
        1. If the user has an exact negation which matches one of the base_scopes,
           or the base_scopes with the verb expanded, we return False.
        2. If the user has an exact scope which matches one of the base_scopes,
           or the base_scopes with the verb expanded, we return True.
        3. If the user has a negation that matches any of the scopes (expanded or not), we return False.
        4. If the user has a scope that matches any of the scopes, we return True.
        5. Return False
    """
    if len(required_scopes) == 0:
        return True

    exclude_exact, include_exact, exclude, include = partition_scopes(granting_scopes)

    required_base_scopes_with_verb = expand_scopes_with_verb(required_scopes, verb)
    required_scopes_with_verb = expand_scopes_with_verb_recursively(
        required_scopes, verb
    )

    # Check case 1
    if any_scope_matches(required_base_scopes_with_verb, exclude_exact):
        return False

    # Check case 2
    if any_scope_matches(required_base_scopes_with_verb, include_exact):
        return True

    # Check case 3
    if any_scope_matches(required_scopes_with_verb, exclude):
        return False

    # Check case 4
    if any_scope_matches(required_scopes_with_verb, include):
        return True

    return False


def any_scope_matches(required_scopes: [str], scopes: [str]):
    """
    Check if any of the given scopes matches any of the required_scopes.
    :param required_scopes:
    :param scopes:
    :return:
    """

    return any(
        scope_matches(strip_negation(required_scope), strip_negation(scope))
        for required_scope in required_scopes
        for scope in scopes
    )


### HELPERS ###
def expand_scopes_with_verb(scopes: [str], verb: str):
    """
    Appends an verb to the scopes given.
    Example:
        scopes: ["user:1", "company:1:user:1"]
        verb: edit
        [
            "user:1:edit",
            "company:1:user:1:edit",
        ]
    :param scopes:
    :param verb:
    :return:
    """
    if not verb:
        return scopes

    return [create_scope(scope, verb) for scope in scopes]


def expand_scopes_with_verb_recursively(scopes: [str], verb: str):
    """
    Appends an verb to all sub-scopes for every scope in a list of scopes.
    Note that the verb itself is also added to the list.
    Example:
        scopes: ["user:1", "company:1:user:1"]
        verb: edit
        [
            "edit",
            "user:1:edit",
            "user:edit",
            "company:1:user:1:edit",
            "company:1:user:edit",
            "company:1:edit",
            "company:edit"
        ]
    :param scopes:
    :param verb:
    :return:
    """
    if not verb:
        return scopes

    result = [verb]
    for scope in scopes:
        parts = scope.split(":")
        for i in range(len(parts)):
            new_scope = create_scope(*parts[: i + 1], verb)
            result.append(new_scope)

    return result


def scope_matches(required_permission: str, granting_permission: str):
    """
    Checks if two scopes match. They match if and only if the following is true:
        - All parts of required_scope are contained in scope, in the same order as supplied in required_scope.
        - If the scope starts with =, it must match the required scope exactly.
    Examples:
        required    = users:1:edit
        scope       = users:1
        OK
        required    = users:1:edit
        scope       = users:1:create
        NOT OK
        required    = users:1
        scope       = users:1:create
        NOT OK
        required    = company:1:timesheets:create
        scope       = company:1
        OK
    :param required_permission:
    :param granting_permission:
    :return:
    """
    if granting_permission[0] == "=":
        return required_permission == granting_permission[1:]

    if granting_permission == required_permission:
        return True

    # Optimisation, bail out when the wildcard is the only permission
    if granting_permission == "*":
        return True

    required_scopes = required_permission.split(":")
    granting_scopes = granting_permission.split(":")

    # A more specified granting scope can never grant access.
    # E.g.
    #  granting = user:1:create
    #  required = user:1
    if len(granting_scopes) > len(required_scopes):
        return False

    if all(
        granting_scope_part == "*"
        or required_scope_part == "*"
        or (granting_scope_part == required_scope_part)
        for required_scope_part, granting_scope_part in zip(
            required_scopes, granting_scopes
        )
    ):
        return True

    return False


def get_scope_arg_str(arg: Union[str, Model, ModelBase]):
    """
    Converts a scope argument into a string.
    For models and modelbases we lookup the model name.
    """
    if isinstance(arg, Model) or isinstance(arg, ModelBase):
        return arg._meta.model_name
    return str(arg)


def create_scope(*args: [Any]):
    """
    Converts a list of scopes into a scope-string.
    :param args:
    :return:
    """
    return ":".join([get_scope_arg_str(arg) for arg in args])


def partition_scopes(scopes: [str]):
    """
    partition_scopes partitions a set of scopes into four sets:
        - Exclude exact
        - Include exact
        - Exclude
        - Include
    """
    exclude_exact = []
    include_exact = []
    exclude = []
    include = []

    for scope in scopes:
        if scope.startswith("-="):
            exclude_exact.append(scope)
        elif scope.startswith("="):
            include_exact.append(scope)
        elif scope.startswith("-"):
            exclude.append(scope)
        else:
            include.append(scope)

    return [exclude_exact, include_exact, exclude, include]


def strip_negation(scope: str) -> str:
    return scope[1:] if scope.startswith("-") else scope



import graphene
from graphene import Node
from graphene_django import DjangoObjectType, DjangoConnectionField

from django_scoped_permissions.models import ScopedPermission


class ScopedPermissionNode(DjangoObjectType):
    class Meta:
        model = ScopedPermission
        interfaces = (Node,)

    @classmethod
    def get_node(self, info, id):
        return ScopedPermission.objects.get(pk=id)


class ScopedPermissionQuery(graphene.ObjectType):
    scoped_permission = Node.Field(ScopedPermissionNode)
    all_scoped_permissions = DjangoConnectionField(ScopedPermissionNode)

    def resolve_all_scoped_permissions(self, info, *args, **kwargs):
        return ScopedPermission.objects.all()



from django.test import TestCase

from django_scoped_permissions.guards import ScopedPermissionGuard


class TestScopedPermissionGuard(TestCase):
    def test__guard_with_scope__grants_and_denies_access_correctly(self):
        guard = ScopedPermissionGuard("scope1:scope2")

        self.assertTrue(guard.has_permission("scope1"))
        self.assertFalse(guard.has_permission("scope2"))
        self.assertTrue(guard.has_permission("scope1:scope2"))
        self.assertFalse(guard.has_permission("scope"))

    def test__guard_with_multiple_scopes__grants_and_denies_access_with_OR_functionality(
        self,
    ):
        guard = ScopedPermissionGuard("scope1:scope2", "scope3:scope4")

        self.assertTrue(guard.has_permission("scope1"))
        self.assertFalse(guard.has_permission("scope2"))
        self.assertTrue(guard.has_permission("scope1:scope2"))
        self.assertFalse(guard.has_permission("scope"))

        self.assertTrue(guard.has_permission("scope3"))
        self.assertFalse(guard.has_permission("scope4"))
        self.assertTrue(guard.has_permission("scope3:scope4"))
        self.assertFalse(guard.has_permission("scope"))

    def test__guard_with_action_and_verb__grants_and_denies_access_correctly(self):
        guard = ScopedPermissionGuard(scope="scope1:scope2", verb="read")

        self.assertTrue(guard.has_permission("scope1:read"))
        self.assertFalse(guard.has_permission("scope2:read"))
        self.assertTrue(guard.has_permission("scope1:scope2:read"))
        self.assertFalse(guard.has_permission("scope:read"))

        self.assertTrue(guard.has_permission("scope1"))
        self.assertFalse(guard.has_permission("scope2"))
        self.assertTrue(guard.has_permission("scope1:scope2"))
        self.assertFalse(guard.has_permission("scope"))

    def test__and_operator__grants_and_denies_acceess_correctly(self):
        guard_one = ScopedPermissionGuard(scope="scope1:scope2", verb="read")
        guard_two = ScopedPermissionGuard(scope="scope3:scope4")

        guard = guard_one & guard_two

        self.assertTrue(guard.has_permission(["scope1:read", "scope3"]))
        self.assertFalse(guard.has_permission(["scope1:read"]))
        self.assertTrue(guard.has_permission(["scope1:scope2:read", "scope3:scope4"]))
        self.assertTrue(guard.has_permission(["scope1", "scope3"]))
        self.assertFalse(guard.has_permission("scope1"))
        self.assertFalse(guard.has_permission("scope2"))

    def test__or_operator__grants_and_denies_acceess_correctly(self):
        guard_one = ScopedPermissionGuard(scope="scope1:scope2", verb="read")
        guard_two = ScopedPermissionGuard(scope="scope3:scope4")

        guard = guard_one | guard_two

        self.assertTrue(guard.has_permission(["scope1:read", "scope3"]))
        self.assertTrue(guard.has_permission(["scope1:read"]))
        self.assertTrue(guard.has_permission(["scope1:scope2:read", "scope3:scope4"]))
        self.assertTrue(guard.has_permission(["scope1", "scope3"]))
        self.assertTrue(guard.has_permission("scope1"))

        self.assertFalse(guard.has_permission("scope4"))
        self.assertFalse(guard.has_permission("scope2"))

    def test__xor_operator__grants_and_denies_acceess_correctly(self):
        guard_one = ScopedPermissionGuard(scope="scope1:scope2", verb="read")
        guard_two = ScopedPermissionGuard(scope="scope3:scope4")

        guard = guard_one ^ guard_two

        self.assertFalse(guard.has_permission(["scope1:read", "scope3"]))
        self.assertTrue(guard.has_permission(["scope1:read"]))
        self.assertFalse(guard.has_permission(["scope1:scope2:read", "scope3:scope4"]))
        self.assertFalse(guard.has_permission(["scope1", "scope3"]))
        self.assertTrue(guard.has_permission("scope1"))

        self.assertFalse(guard.has_permission("scope4"))
        self.assertFalse(guard.has_permission("scope2"))

    def test__not_operator__grants_and_denies_access_correctly(self):
        guard_one = ScopedPermissionGuard(scope="scope1:scope2", verb="read")

        guard = ~guard_one

        self.assertFalse(guard.has_permission("scope1:read"))
        self.assertTrue(guard.has_permission("scope2:read"))
        self.assertFalse(guard.has_permission("scope1:scope2:read"))
        self.assertTrue(guard.has_permission("scope:read"))

        self.assertFalse(guard.has_permission("scope1"))
        self.assertTrue(guard.has_permission("scope2"))
        self.assertFalse(guard.has_permission("scope1:scope2"))
        self.assertTrue(guard.has_permission("scope"))

    def test__guard_with_context__is_expanded_and_checked(self):
        guard = ScopedPermissionGuard("scope1:{var}")

        self.assertTrue(guard.has_permission("scope1", context={}))
        self.assertTrue(
            guard.has_permission("scope1:scope2", context={"var": "scope2"})
        )
        self.assertTrue(
            guard.has_permission(
                "scope1:scope3",
                context={"var": ["scope2", "scope3"], "other": ["scope4"]},
            )
        )

    def test__guard_with_empty_list__returns_always_true_guard(self):
        guard = ScopedPermissionGuard([])
        self.assertTrue(guard.has_permission("something"))
        self.assertTrue(guard.has_permission("something:else"))
        self.assertTrue(guard.has_permission("anything?"))



from typing import Tuple, Mapping, Union, Iterable, List

from graphene import Node
from graphene_django import DjangoObjectType
from graphene_django.types import DjangoObjectTypeOptions
from graphene_django_cud.mutations import (
    DjangoPatchMutation,
    DjangoDeleteMutation,
    DjangoBatchCreateMutation,
    DjangoBatchDeleteMutation,
    DjangoFilterDeleteMutation,
)
from graphene_django_cud.mutations.create import DjangoCreateMutation
from graphene_django_cud.mutations.delete import DjangoDeleteMutationOptions
from graphene_django_cud.mutations.patch import DjangoPatchMutationOptions
from graphene_django_cud.mutations.update import (
    DjangoUpdateMutationOptions,
    DjangoUpdateMutation,
)
from graphql import GraphQLError

from django_scoped_permissions.guards import ScopedPermissionGuard
from django_scoped_permissions.models import (
    ScopedModelMixin,
    ScopedPermissionHolderMixin,
)
from django_scoped_permissions.util import (
    create_resolver_from_method,
    create_resolver_from_scopes,
)


class ScopedDjangoNodeOptions(DjangoObjectTypeOptions):
    allow_anonymous = False  # type: bool
    node_permissions = None  # type: Tuple[str]
    field_permissions = None  # type: Mapping[str, Union[bool, Tuple[str]]]
    verb = "read"  # type: str


class ScopedDjangoNode(DjangoObjectType):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        node_permissions=None,
        field_permissions=None,
        allow_anonymous=False,
        verb="read",
        _meta=None,
        **options,
    ):
        if not _meta:
            _meta = ScopedDjangoNodeOptions(cls)

        _meta.allow_anonymous = allow_anonymous
        _meta.node_permissions = node_permissions
        _meta.field_permissions = field_permissions
        _meta.verb = verb
        interfaces = options.get("interfaces", ()) + (Node,)

        # Great, the class is set up. Now let's add permission guards.
        field_permissions = field_permissions or {}

        permission_guard = ScopedPermissionGuard(node_permissions)
        _meta.permission_guard = permission_guard

        super().__init_subclass_with_meta__(
            _meta=_meta, interfaces=interfaces, **options
        )

        for field, permissions in field_permissions.items():
            if callable(permissions):
                if hasattr(cls, f"resolve_{field}"):
                    continue

                setattr(
                    cls,
                    f"resolve_{field}",
                    create_resolver_from_method(field, permissions),
                )
            elif isinstance(permissions, tuple) or isinstance(permissions, list):
                if hasattr(cls, f"resolve_{field}"):
                    continue

                setattr(
                    cls,
                    f"resolve_{field}",
                    create_resolver_from_scopes(field, permissions),
                )
            elif isinstance(permissions, str):
                if hasattr(cls, f"resolve_{field}"):
                    continue

                setattr(
                    cls,
                    f"resolve_{field}",
                    create_resolver_from_scopes(field, [permissions]),
                )
            else:
                raise ValueError(
                    f"Invalid field type {type(permissions)} given to ScopedDjangoNode for field {field}"
                )

    @classmethod
    def get_node(cls, info, id):
        user = info.context.user
        if not cls._meta.allow_anonymous and not isinstance(
            user, ScopedPermissionHolderMixin
        ):
            raise GraphQLError("You are not permitted to view this.")

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        Model = cls._meta.model
        obj = Model.objects.get(pk=id)

        context = {
            "user": user,
            "context": info.context,
            # Deprecated
            "obj": obj,
        }

        if isinstance(obj, ScopedModelMixin):
            context["base_scopes"] = obj.get_base_scopes()
            context["required_scopes"] = obj.get_required_scopes()

        # If we have explicit permission, we check against the guard
        if cls._meta.node_permissions:
            if not cls._meta.permission_guard.has_permission(
                granting_permissions, context
            ):
                raise GraphQLError("You are not permitted to view this.")
        elif isinstance(obj, ScopedModelMixin):
            if not isinstance(user, ScopedPermissionHolderMixin):
                raise GraphQLError("You are not permitted to view this.")

            if not obj.can_be_accessed_by(user, cls._meta.verb):
                raise GraphQLError("You are not permitted to view this.")

        return super().get_node(info, id)


class ScopedDjangoCreateMutation(DjangoCreateMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = cls.get_permissions(root, info, input) or []

        user = info.context.user

        permission_guard = ScopedPermissionGuard(permissions)
        context = {"context": info.context, "input": input, "user": info.context.user}

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        if not permission_guard.has_permission(granting_permissions, context=context):
            raise GraphQLError("You are not permitted to view this.")


class ScopedDjangoBatchCreateMutation(DjangoBatchCreateMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = cls.get_permissions(root, info, input) or []

        if not hasattr(permissions, "__len__") or len(permissions) == 0:
            return

        user = info.context.user

        permission_guard = ScopedPermissionGuard(permissions)
        context = {"context": info.context, "input": input, "user": info.context.user}

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        if not permission_guard.has_permission(granting_permissions, context=context):
            raise GraphQLError("You are not permitted to view this.")


class ScopedDjangoPatchMutationOptions(DjangoPatchMutationOptions):
    verb = "update"  # type: str


class ScopedDjangoPatchMutation(DjangoPatchMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_permissions(
        cls, root, info, input, id, obj
    ) -> Union[Iterable[str], ScopedPermissionGuard]:
        super_permissions = super().get_permissions(root, info, input, id, obj) or []

        if (
            hasattr(super_permissions, "__len__")
            and len(super_permissions) > 0
            or isinstance(super_permissions, ScopedPermissionGuard)
        ):
            return super_permissions

        # If we don't have explicit permissions we use some defaults here.
        return ScopedPermissionGuard(scope="{required_scopes}", verb=cls._meta.verb)

    @classmethod
    def check_permissions(cls, root, info, input, id, obj) -> None:
        permissions = cls.get_permissions(root, info, input, id, obj) or []

        if not hasattr(permissions, "__len__") or len(permissions) == 0:
            return

        user = info.context.user

        permission_guard = ScopedPermissionGuard(permissions)
        context = {}

        if isinstance(obj, ScopedModelMixin):
            context["base_scopes"] = obj.get_base_scopes()
            context["required_scopes"] = obj.get_required_scopes()

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        context["context"] = info.context
        context["input"] = input
        context["id"] = id
        context["obj"] = obj

        if not permission_guard.has_permission(granting_permissions, context=context):
            raise GraphQLError("You are not permitted to view this.")

    @classmethod
    def __init_subclass_with_meta__(cls, _meta=None, verb="update", **options):
        if _meta is None:
            _meta = ScopedDjangoPatchMutationOptions(cls)

        _meta.verb = verb

        return super().__init_subclass_with_meta__(_meta=_meta, **options)


class ScopedDjangoUpdateMutationOptions(DjangoUpdateMutationOptions):
    verb = "update"  # type: str


class ScopedDjangoUpdateMutation(DjangoUpdateMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_permissions(
        cls, root, info, input, id, obj
    ) -> Union[Iterable[str], ScopedPermissionGuard]:
        super_permissions = super().get_permissions(root, info, input, id, obj) or []

        if (
            hasattr(super_permissions, "__len__")
            and len(super_permissions) > 0
            or isinstance(super_permissions, ScopedPermissionGuard)
        ):
            return super_permissions

        # If we don't have explicit permissions we use some defaults here.
        return ScopedPermissionGuard(scope="{required_scopes}", verb=cls._meta.verb)

    @classmethod
    def check_permissions(cls, root, info, input, id, obj) -> None:
        permissions = cls.get_permissions(root, info, input, id, obj) or []

        if not hasattr(permissions, "__len__") or len(permissions) == 0:
            return

        user = info.context.user

        permission_guard = ScopedPermissionGuard(permissions)
        context = {}

        if isinstance(obj, ScopedModelMixin):
            context["base_scopes"] = obj.get_base_scopes()
            context["required_scopes"] = obj.get_required_scopes()

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        context["context"] = info.context
        context["input"] = input
        context["id"] = id
        context["obj"] = obj

        if not permission_guard.has_permission(granting_permissions, context=context):
            raise GraphQLError("You are not permitted to view this.")

    @classmethod
    def __init_subclass_with_meta__(cls, _meta=None, verb="update", **options):
        if _meta is None:
            _meta = ScopedDjangoUpdateMutationOptions(cls)

        _meta.verb = verb

        return super().__init_subclass_with_meta__(_meta=_meta, **options)


class ScopedDjangoDeleteMutationOptions(DjangoDeleteMutationOptions):
    verb = "delete"  # type: str


class ScopedDjangoDeleteMutation(DjangoDeleteMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_permissions(
        cls, root, info, id, obj
    ) -> Union[Iterable[str], ScopedPermissionGuard]:
        super_permissions = super().get_permissions(root, info, id, obj) or []

        if (
            hasattr(super_permissions, "__len__")
            and len(super_permissions) > 0
            or isinstance(super_permissions, ScopedPermissionGuard)
        ):
            return super_permissions

        return ScopedPermissionGuard(scope="{required_scopes}", verb=cls._meta.verb)

    @classmethod
    def check_permissions(cls, root, info, id, obj) -> None:
        permissions = cls.get_permissions(root, info, id, obj) or []

        if not hasattr(permissions, "__len__") or len(permissions) == 0:
            return

        user = info.context.user

        permission_guard = ScopedPermissionGuard(permissions)
        context = {}

        if isinstance(obj, ScopedModelMixin):
            context["base_scopes"] = obj.get_base_scopes()
            context["required_scopes"] = obj.get_required_scopes()

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        context["context"] = info.context
        context["id"] = id
        context["obj"] = obj

        if not permission_guard.has_permission(granting_permissions, context=context):
            raise GraphQLError("You are not permitted to view this.")

    @classmethod
    def __init_subclass_with_meta__(cls, _meta=None, verb="delete", **options):
        if _meta is None:
            _meta = ScopedDjangoUpdateMutationOptions(cls)

        _meta.verb = verb

        return super().__init_subclass_with_meta__(_meta=_meta, **options)


class ScopedDjangoBatchDeleteMutation(DjangoBatchDeleteMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = cls.get_permissions(root, info, input) or []

        if not hasattr(permissions, "__len__") or len(permissions) == 0:
            return

        user = info.context.user

        permission_guard = ScopedPermissionGuard(permissions)
        context = {"context": info.context, "input": input, "user": info.context.user}

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        if not permission_guard.has_permission(granting_permissions, context=context):
            raise GraphQLError("You are not permitted to view this.")


class ScopedDjangoFilterDeleteMutation(DjangoFilterDeleteMutation):
    class Meta:
        abstract = True

    @classmethod
    def check_permissions(cls, root, info, input) -> None:
        permissions = cls.get_permissions(root, info, input) or []

        if not hasattr(permissions, "__len__") or len(permissions) == 0:
            return

        user = info.context.user

        permission_guard = ScopedPermissionGuard(permissions)
        context = {"context": info.context, "input": input, "user": info.context.user}

        granting_permissions = (
            user.get_granting_scopes() if hasattr(user, "get_granting_scopes") else []
        )

        if not permission_guard.has_permission(granting_permissions, context=context):
            raise GraphQLError("You are not permitted to view this.")



from addict import Dict
from django.core.exceptions import PermissionDenied
from django.test import TestCase

from django_scoped_permissions.decorators import gql_has_scoped_permissions
from django_scoped_permissions.guards import ScopedPermissionGuard
from django_scoped_permissions.models import ScopedPermission
from django_scoped_permissions.tests.factories import UserFactory, CompanyFactory


class TestGqlHasScopedPermissions(TestCase):
    def test__simple_scope_check_with_permission__permits(self):
        @gql_has_scoped_permissions("scope1:scope2")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope1:scope2")
        user.scoped_permissions.add(perm)
        info = Dict()
        info.context.user = user
        wrapper_method(None, info)

    def test__simple_scope_check_without_permission__fails(self):
        @gql_has_scoped_permissions("scope1:scope2")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope3")
        user.scoped_permissions.add(perm)
        info = Dict()
        info.context.user = user
        with self.assertRaises(PermissionDenied):
            wrapper_method(None, info)

    def test_expand_from_context__check_with_organization__succeeds(self):
        @gql_has_scoped_permissions("company:{context.company.id}:read")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="company:1")
        user.scoped_permissions.add(perm)
        info = Dict()
        info.context.user = user
        info.context.company = CompanyFactory.create()
        wrapper_method(None, info)

    def test__multiple_permisions__becomes_or_statement_by_default(self):
        @gql_has_scoped_permissions("scope1", "scope2")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope1")
        user.scoped_permissions.add(perm)

        info = Dict()
        info.context.user = user
        info.context.company = CompanyFactory.create()
        wrapper_method(None, info)

        user.scoped_permissions.all().delete()
        perm = ScopedPermission.objects.create(scope="scope2")
        user.scoped_permissions.add(perm)
        wrapper_method(None, info)

        with self.assertRaises(PermissionDenied):
            user.scoped_permissions.all().delete()
            perm = ScopedPermission.objects.create(scope="scope3")
            user.scoped_permissions.add(perm)
            wrapper_method(None, info)

    def test__with_guard__uses_guard(self):
        @gql_has_scoped_permissions(ScopedPermissionGuard("scope1"))
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope1")
        user.scoped_permissions.add(perm)

        info = Dict()
        info.context.user = user
        info.context.company = CompanyFactory.create()
        wrapper_method(None, info)

    def test__with_verb_and_scope__uses_verb_and_scope(self):
        @gql_has_scoped_permissions(scope="scope1", verb="verb")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope1:verb")
        user.scoped_permissions.add(perm)

        info = Dict()
        info.context.user = user
        info.context.company = CompanyFactory.create()
        wrapper_method(None, info)

        user.scoped_permissions.all().delete()
        perm = ScopedPermission.objects.create(scope="verb")
        user.scoped_permissions.add(perm)
        wrapper_method(None, info)

    def test__with_verb_and_scope__uses_verb_and_scope(self):
        @gql_has_scoped_permissions(scope="scope1", verb="verb")
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope1:verb")
        user.scoped_permissions.add(perm)

        info = Dict()
        info.context.user = user
        info.context.company = CompanyFactory.create()
        wrapper_method(None, info)

        user.scoped_permissions.all().delete()
        perm = ScopedPermission.objects.create(scope="verb")
        user.scoped_permissions.add(perm)
        wrapper_method(None, info)

    def test__complex_permission_guard_combination__uses_guard(self):
        # Matches when a user has scope1 and not scope2:read (or just read) OR
        # when the user has scope2:scope3 xor scope4
        @gql_has_scoped_permissions(
            (
                ScopedPermissionGuard("scope1")
                & ~ScopedPermissionGuard(scope="scope2", verb="read")
            )
            | (ScopedPermissionGuard("scope2:scope3") ^ ScopedPermissionGuard("scope4"))
        )
        def wrapper_method(self, info):
            pass

        user = UserFactory.create()
        perm = ScopedPermission.objects.create(scope="scope1")
        user.scoped_permissions.add(perm)

        info = Dict()
        info.context.user = user
        info.context.company = CompanyFactory.create()
        wrapper_method(None, info)

        user.scoped_permissions.all().delete()
        perm = ScopedPermission.objects.create(scope="scope2:scope3")
        user.scoped_permissions.add(perm)
        wrapper_method(None, info)

        user.scoped_permissions.all().delete()
        perm = ScopedPermission.objects.create(scope="scope4")
        user.scoped_permissions.add(perm)
        wrapper_method(None, info)

        with self.assertRaises(PermissionDenied):
            user.scoped_permissions.all().delete()
            perm = ScopedPermission.objects.create(scope="scope1")
            perm = ScopedPermission.objects.create(scope="scope2:read")
            user.scoped_permissions.add(perm)
            wrapper_method(None, info)

        with self.assertRaises(PermissionDenied):
            user.scoped_permissions.all().delete()
            user.add_or_create_permission("scope1")
            user.add_or_create_permission("read")
            wrapper_method(None, info)



from typing import Optional

from django.db import models
from django.db.models import Value, F, Case, When
from django.db.models.functions import Concat

from django_scoped_permissions.core import any_scope_matches, scopes_grant_permissions


class ScopedPermission(models.Model):
    class Meta:
        unique_together = (("scope", "exclude", "exact"),)

    scope = models.TextField(blank=False)
    exclude = models.BooleanField(
        default=False,
        help_text="Whether this should be an exclusive permission, meaning that if scope is 'user:update' and exclude is True then users with this usertype will not be able to update users, even their own.",
    )
    exact = models.BooleanField(
        default=False,
        help_text="If checked, the permission needs an exact match to count. In other words, it does not work recursively as standard scoped permissions.",
    )

    def get_scope_parts(self):
        return self.scope.split(":")

    def __str__(self):
        prefix = "-" if self.exclude else ""
        prefix += "=" if self.exact else ""
        return prefix + self.scope


class ScopedPermissionGroup(models.Model):
    name = models.TextField()
    scoped_permissions = models.ManyToManyField(
        ScopedPermission, blank=True, related_name="in_groups"
    )

    def __str__(self):
        return self.name


class ScopedPermissionHolderMixin:
    def get_granting_scopes(self):
        return []

    def has_scoped_permissions(self, *required_scopes):
        return self.has_any_scoped_permissions(*required_scopes)

    def has_any_scoped_permissions(self, *required_scopes):
        scopes = self.get_granting_scopes()

        return scopes_grant_permissions(required_scopes, scopes)

    def has_all_scoped_permissions(self, *required_scopes):
        scopes = self.get_granting_scopes()

        for scope in required_scopes:
            if not any_scope_matches([scope], scopes):
                return False

        return True

    def has_access_to(self, model: "ScopedModelMixin", verb: Optional[str] = None):

        granting_scopes = self.get_granting_scopes()
        required_scopes = model.get_required_scopes()

        return scopes_grant_permissions(required_scopes, granting_scopes, verb)


class ScopedPermissionHolder(models.Model, ScopedPermissionHolderMixin):
    class Meta:
        abstract = True

    scoped_permissions = models.ManyToManyField(ScopedPermission, blank=True)
    scoped_permission_groups = models.ManyToManyField(ScopedPermissionGroup, blank=True)

    @property
    def resolved_group_scopes(self):
        scopes = ScopedPermission.objects.filter(
            in_groups__in=self.scoped_permission_groups.all()
        )
        scopes = scopes.annotate(
            parsed_scope=Concat(
                Case(When(exclude=True, then=Value("-")), default=Value("")),
                Case(When(exact=True, then=Value("=")), default=Value("")),
                F("scope"),
            )
        )

        return list(scopes.values_list("parsed_scope", flat=True))

    @property
    def resolved_scopes(self):
        scopes = self.scoped_permissions.all() | ScopedPermission.objects.filter(
            in_groups__in=self.scoped_permission_groups.all()
        )
        scopes = scopes.annotate(
            parsed_scope=Concat(
                Case(When(exclude=True, then=Value("-")), default=Value("")),
                Case(When(exact=True, then=Value("=")), default=Value("")),
                F("scope"),
            )
        )

        resolved_scopes = list(scopes.values_list("parsed_scope", flat=True))

        return resolved_scopes

    def get_scopes(self):
        """
        DEPRECATED: Use `get_granting_scopes` instead
        """
        return self.resolved_scopes

    def get_granting_scopes(self):
        return self.resolved_scopes

    def has_scoped_permissions(self, *required_scopes):
        return self.has_any_scoped_permissions(*required_scopes)

    def has_any_scoped_permissions(self, *required_scopes):
        scopes = self.get_granting_scopes()

        return scopes_grant_permissions(required_scopes, scopes)

    def has_all_scoped_permissions(self, *required_scopes):
        scopes = self.get_granting_scopes()

        for scope in required_scopes:
            if not any_scope_matches([scope], scopes):
                return False

        return True

    def add_or_create_permission(
        self, scoped_permission: str, exact=False, exclude=False
    ):
        """
        Helper method which adds a permission defined by the arguments to the permission holder.
        If a ScopedPermission object matching the argument does not exist, a new one will be created.

        The exact and exclude properties of the scope can be added either directly in the string, e.g.

            add_or_create_permission("-=scope1:scope2")

        Or as parameters

            add_or_create_permission("scope1:scope2", exact=False, exclude=False)
        """

        if scoped_permission.startswith("-"):
            exclude = True
            scoped_permission = scoped_permission[1:]

        if scoped_permission.startswith("="):
            exact = True
            scoped_permission = scoped_permission[1:]

        scope, _ = ScopedPermission.objects.get_or_create(
            scope=scoped_permission, exclude=exclude, exact=exact
        )

        self.scoped_permissions.add(scope)


# DEPRECATED: Use ScopedPermissionHolder
HasScopedPermissionMixin = ScopedPermissionHolder


class ScopedModelMixin:
    def get_base_scopes(self):
        """
        DEPRECATED: Use `get_required_scopes`
        """
        return self.get_required_scopes()

    def get_required_scopes(self):
        return []

    def can_be_accessed_by(
        self, holder: ScopedPermissionHolderMixin, verb: Optional[str] = None
    ):
        user_scopes = holder.get_granting_scopes()
        required_scopes = self.get_required_scopes()

        return scopes_grant_permissions(required_scopes, user_scopes, verb)

    def has_permission(
        self, user: ScopedPermissionHolderMixin, verb: Optional[str] = None
    ):
        """
        DEPRECATED: Use `can_be_accessed_by`.
        """
        return self.can_be_accessed_by(user, verb)


class ScopedModel(models.Model, ScopedModelMixin):
    class Meta:
        abstract = True



from django.test import TestCase

from django_scoped_permissions.core import scope_matches, scope_grants_permission


class TestScopeGrantsPermission(TestCase):
    def test__equal_scopes__returns_true(self):
        self.assertTrue(
            scope_grants_permission("user:1", "user:1")
        )

    def test__required_scope_starts_with_granting_scope__returns_true(self):
        self.assertTrue(
            scope_grants_permission("user:1:edit", "user:1")
        )
        self.assertTrue(
            scope_grants_permission("organization:edit:user", "organization:edit")
        )

    def test__required_scope_ends_with_granting_scope__returns_false(self):
        self.assertFalse(
            scope_grants_permission("organization:user:1", "user:1")
        )

    def test__granting_scope_is_substring_but_does_not_match_parent_scope__returns_false(self):
        self.assertTrue(
            scope_grants_permission("user:1:edit", "user")
        )
        self.assertFalse(
            scope_grants_permission("organization:edit:user", "organization:ed")
        )

    def test__exact_scope_matches_only_on_exact_match(self):
        self.assertTrue(
            scope_grants_permission("user:1", "=user:1")
        )
        self.assertFalse(
            scope_grants_permission("user:1:update", "=user:1")
        )

    def test__exclusive_scope__always_fails(self):
        self.assertFalse(
            scope_grants_permission("user:1", "-user:1")
        )
        self.assertFalse(
            scope_grants_permission("user:1:subscope", "-user:1")
        )
        self.assertFalse(
            scope_grants_permission("user:1", "-read", "read")
        )


class TestScopeMatches(TestCase):
    def test__equal_scopes__returns_true(self):
        self.assertTrue(
            scope_matches("user:1", "user:1")
        )

    def test__required_scope_starts_with_granting_scope__returns_true(self):
        self.assertTrue(
            scope_matches("user:1:edit", "user:1")
        )
        self.assertTrue(
            scope_matches("organization:edit:user", "organization:edit")
        )

    def test__required_scope_ends_with_granting_scope__returns_false(self):
        self.assertFalse(
            scope_matches("organization:user:1", "user:1")
        )

    def test__granting_scope_is_substring_but_does_not_match_parent_scope__returns_false(self):
        self.assertFalse(
            scope_matches("user:1:edit", "use")
        )
        self.assertFalse(
            scope_matches("organization:edit:user", "organization:ed")
        )

    def test__wildcard_on_granting_scope__matches(self):
        self.assertTrue(
            scope_matches("user:1:edit", "*")
        )
        self.assertTrue(
            scope_matches("organization:edit:user", "*")
        )
        self.assertTrue(
            scope_matches("organization:edit:user", "organization:*")
        )
        self.assertTrue(
            scope_matches("organization:edit:user", "organization:*:user")
        )
        self.assertTrue(
            scope_matches("organization:edit:user", "*:edit:user")
        )
        self.assertTrue(
            scope_matches("organization:edit:user", "organization:*:user")
        )
        self.assertFalse(
            scope_matches("organization:edit:user", "organization:edit:user:*")
        )
        self.assertFalse(
            scope_matches("organization:edit:user", "*:edit:user:*")
        )

    def test__wildcard_on_required__scope__matches(self):
        self.assertTrue(
            scope_matches("user:*:edit", "user:1:edit")
        )
        self.assertTrue(
            scope_matches("organization:*:user", "organization:4:user")
        )
        self.assertTrue(
            scope_matches("organization:*", "organization:clients")
        )
        self.assertFalse(
            scope_matches("organization:*", "organization:clients:read")
        )
